(*--------------------------------------------------------------------------*)
(*  Santronics Software, Co.                                                *)
(*  30034 SW 153 Ct.                                                        *)
(*  Leisure City, FL  33033                                                 *)
(*                                                                          *)
(*                                                                          *)
(*             (C) Copyright 1988-95, Santronics Software, Co               *)
(*                                                                          *)
(*--------------------------------------------------------------------------*)

(*

   These functions are used to read/write the users selected areas into
   the XPUSER.XUA.  This is a very complicated logic and requires alot
   of knowledge about how Xpress Works.

   Basically, Xpress stores a record XUA_TYPE for every forum selected for
   the user.  If the BBS uses an ECHO TAG system, then the ADX system (see
   below) is used to use the CRC of the echo tag as an INDEX to the record.
   Otherwise, Xpress uses the AREA NUMBER as the index.

   The functions below, allow you to read a BIT MAP of areas selected. The
   bit map array is 4K (SetMarray) with each byte holding 8 bits, this
   allows for 32 thousands conferences.

   So the idea is to read each record in XPUSER.XUA, check the user crc to
   see if this record belongs to him.  If the record belongs to him, use
   the areanumber and bit map it into the XUA_SET global array.  You will
   also see there is a SYSOP SELECTED bit flag.  If this bit is set, then
   map the area number into the xsa_set global array.

*)

(****************************************************************************

  ADX COMPILER DIRECTIVE.  If you are using Maximus or OPUS, you need to use
  the ADX compiler directive string to utilize the XPAREAS.ADX file

  Maximus
  Opus 1.20
  Opus 1.73

  Explaination of XPAREAS.ADX:

  The XPAREAS.ADX is special ECHO TAG crc index file used by the Silver
  Xpress Mail Door.  It was designed to help those systems (specifically
  Maximus) where there was no AREA NUMBER system.

  It is not needed by 3rd party utility, but if you want to know what
  it does here is a short discussion.

  The XPAREAS.ADX is created by XPADM to produce a CRC index into the
  main forum file, XPAREAS.DAT.  The mail door will send this file
  XPAREAS.ADX to the user with the mail packet.

  When mail is uploaded, Xpress will check the area number which the
  uploaded mail says to toss it into and grab the ECHO CRC from a COPY of
  XPAREAS.ADX provided by the reader.

  The ECHO CRC is then searched in the current XPAREAS.ADX for the BBS
  and if found, the AREA NUMBER in the current XPAREAS.ADX is used.

  The reason this is done is to assure we can find the right area number
  for situations where the sysop might had changes areas around.  By giving
  the user a copy of the XPAREAS.ADX, Xpress will have a history of how the
  areas might have changed.

****************************************************************************)

{$DEFINE ADX}

{$O+,R-,A-,I-}

unit xuatools;

interface

uses

     control,
     xpmarea,
     bitset,
     crt,
     dos;


Type XUA_TYPE =  (* xpress user area record structure *)
     record
       status     : integer;   (* -1 unused recorded            *)
       usrcrc     : longint;   (* user's 32 bit crc             *)
       areanum    : word;      (* actual area number. For MAX it's the INDEX *)
       scanlimit  : word;      (* user defined scan limit       *)
       xuaflags   : word;      (* see XUA_FLAGS                 *)
       areacrc    : longint;   (* CRC of ECHO NAME              *)
     end;
    fv_xuatype    = file of xua_type;


(* bit flags for xuaflags *)

Const
     _XUA_SYSOPSELECTED  = $0001;
     _XUA_NOBODYSEARCH   = $0002;       (* not used, future  *)
     _XUA_DIRECTONLY     = $0004;       (* not used, future  *)


Var  fv_xua          : fv_xuatype;
     xua_rec         : xua_type;


     xua_set         : SetMType;  (* Set of Active areas for user           *)
     xsa_set         : SetMType;  (* Set of sysop selected areas for user   *)

                                  (* For Max, it's the INDEX of the area    *)
                                  (* For others, it's the actual area       *)
                                  (* But it shouldn't matter because the    *)
                                  (* the area crc is check agains the user's*)
                                  (* selected areas crc.                    *)

Const
     HIGHEST_AREANUM : INTEGER   = 0;
{$IFDEF ADX}
     USE_ADX         : BOOLEAN   = TRUE;
{$ELSE}
     USE_ADX         : BOOLEAN   = FALSE;
{$ENDIF}

Procedure Delete_XuaFile_By_Mem(path : pathstr; ucrc:longint);
Procedure open_xua_file(path : pathstr; shareon : boolean);
Procedure read_xuafile_to_xuaset_by_file(path : pathstr; ucrc:longint; shareon : boolean);
Procedure read_xuafile_to_xuaset_by_mem(path : pathstr; ucrc:longint; shareon : boolean);

(*
   These routines are practicvally the same. The difference is one uses
   the XUA_SET while the other uses msgres record. Logic can be added to
   combine these routines.
 *)

Procedure save_xuaset_to_xuafile_by_mem(path : pathstr; ucrc:longint; shareon : boolean);
Procedure save_xuaset_to_xuafile_by_file(path : pathstr; ucrc:longint; shareon : boolean);

{$IFDEF ADX}
  type xpadxptrtype = array[0..0] of _xpaidx;

  Const
       totadx   : word = 0;
       xpadxptr : ^xpadxptrtype = nil;
  var  xpaidx   : _xpaidx;
  Procedure GetXpAdxEchoCrc(aidx:word;var xpadx:_xpaidx);
  function FindXpAdxByArea(area : word; var xpadx:_xpaidx): Boolean;
  Procedure OpenXpAdxFile(fn : pathstr);
  Procedure FindXpAdxCrc(crc:longint; var ia : integer);
  Procedure FindXpAdxCrcByArea(var crc:longint; ia : integer);
  Function FindXpAdxAreaByCrc(crc:longint; var a : integer) : boolean;
  Procedure CloseXpAdxFile;
{$ENDIF}

implementation

Const LockRegion   = 00;
      UnLockRegion = 01;

Function FileLock(handle : word; action : byte; start,bytes : longint; var ax : integer): Boolean;
  var reg : registers;

   begin
    reg.ax     := $5C00 + action;
    reg.bx     := handle;
    reg.cx     := hi(start);
    reg.dx     := lo(start);
    reg.si     := hi(bytes);
    reg.di     := lo(bytes);
    reg.flags  := 0;
    Intr($21,reg);
    FileLock := (reg.flags and $01) = $00;
    ax := reg.ax;
   end;

{$IFDEF ADX}

  Procedure GetXpAdxEchoCrc(aidx:word;var xpadx:_xpaidx);
     begin
       fillchar(xpadx,sizeof(xpadx),0);
       if (xpadxptr = nil) or (totadx = 0)  then exit;
       if (aidx > (totadx-1)) then exit;
       xpadx := xpadxptr^[aidx];
     end;

  Procedure FindXpAdxCrc(crc:longint; var ia : integer);
     begin
       ia := -1;
       if (xpadxptr = nil) or (totadx = 0)  then exit;
       ia := 0;
       while (ia < totadx) and (xpadxptr^[ia].echo_crc <> crc) do Inc(ia);
       if ia = totadx then ia := -1;
     end;

  Function FindXpAdxAreaByCrc(crc:longint; var a : integer) : boolean;
  var ia : integer;
     begin
       FindXpAdxAreaByCrc := FALSE;
       if (xpadxptr = nil) or (totadx = 0)  then exit;
       ia := 0;
       while (ia < totadx) and (xpadxptr^[ia].echo_crc <> crc) do Inc(ia);
       if ia = totadx then ia := -1;
       if ia <> -1 then
          begin
           FindXpAdxAreaByCrc := TRUE;
           a := xpadxptr^[ia].area;
          end;
     end;

 function FindXpAdxByArea(area : word; var xpadx:_xpaidx): Boolean;
  var i : word;
     begin
       FindXpadxByArea := FALSE;
       if (xpadxptr = nil) or (totadx = 0)  then exit;
       i := 0;
       while (i < totadx) and (xpadxptr^[i].area <> area) do Inc(i);
       if i < totadx then
          begin
            FindXpadxByArea := TRUE;
            xpadx := xpadxptr^[i];
          end;
     end;

  Procedure FindXpAdxCrcByArea(var crc:longint; ia : integer);
  var i : word;
     begin
       crc := -1;
       if (xpadxptr = nil) or (totadx = 0)  then exit;
       i := 0;
       while (i < totadx) and (xpadxptr^[i].area <> ia) do Inc(i);
       if i < totadx then crc := xpadxptr^[i].echo_crc;
     end;

  Procedure OpenXpAdxFile(fn : pathstr);
  var fv : file;
      fs : longint;
      ab : word;
      fm : byte;
     begin
       fm := filemode;
       fm := $40;
       assign(fv,fn);
       reset(fv,1);
       fm := filemode;
       if ioresult <> 0 then exit;
       fs     := filesize(fv);
       totadx := fs div sizeof(_xpaidx);
       xpadxptr := nil;
       if fs <> 0 then
          begin
            getmem(xpadxptr,fs);
            blockread(fv,xpadxptr^[0],fs,ab);
          end;
       close(fv);
     end;

  Procedure CloseXpAdxFile;
     begin
       if xpadxptr <> nil then
          begin
            getmem(xpadxptr,totadx * sizeof(_xpaidx));
            xpadxptr := nil;
            totadx   := 0;
          end;
     end;

{$ENDIF}


procedure open_xua_file(path:pathstr; shareon : boolean);
var f : byte; cnt : word; err : integer; lk : boolean;
  begin
   assign(fv_xua,path+xua_file_name);
   cnt := 3000;
   f   := FileMode;
   if shareon then filemode := $42;  (* readwrite/denynone *)
   repeat
     reset(fv_xua);
     err := ioresult;
     case err of
       0  : ;                        (* ok        *)
       2  : rewrite(fv_xua);         (* create it *)
       5,
      167 : begin
             delay(10);
             dec(cnt);
            end;
      else  writeln('XUAFILE ERROR : ',ERR);
     end;
   until (err=0) or (cnt=0);
   filemode := f;
  end;

Procedure read_xuafile_to_xuaset_by_file(path : pathstr; ucrc:longint; shareon : boolean);
{ read XPUSER.XUA file }
var ia : integer;
  begin
   open_xua_file(path+xua_file_name,shareon);
{$IFDEF ADX}
   OpenXpAdxFile(path+adx_file_name);
{$ENDIF}
   while not eof(fv_xua) do
     begin
      read(fv_xua,xua_rec);
      if integer(HIGHEST_AREANUM) < xua_rec.areanum then
         HIGHEST_AREANUM := xua_rec.areanum;
      if xua_rec.status <> -1 then
         if xua_rec.usrcrc = ucrc then
            begin
{$IFDEF ADX}
                  ia := xua_rec.areanum;
                  if Use_ADX then
                     begin
                       if not FindXpAdxAreaByCrc(xua_rec.areacrc,ia) then
                          begin
                            ia := -1;
                          end;
                     end;
                  if ia <> -1 then
                     begin
                       SetAreaBit(xua_set,ia);
                       if IsBitSet(_XUA_SYSOPSELECTED,xua_rec.xuaflags) then
                             SetAreabit(xsa_set,ia);
                     end;
{$ELSE}
                  SetAreaBit(xua_set,xua_rec.areanum);
                  if IsBitSet(_XUA_SYSOPSELECTED,xua_rec.xuaflags) then
                        SetAreabit(xsa_set,xua_rec.areanum);
{$ENDIF}
            end;
     end;
   close(fv_xua);
{$IFDEF ADX}
   CloseXpADXFile;
{$ENDIF}
  end;

type xuabuftype = array[1..1] of xua_type;

Procedure read_xuafile_to_xuaset_by_mem(path : pathstr; ucrc:longint; shareon : boolean);
{ read XPUSER.XUA file }
var buf     : ^xuabuftype;
    recs    : longint;
    ab      : word;
    fv      : file;
    ia      : integer;
  begin
   ab := filemode;
   filemode := $40; {_readonly + _denynone}
   assign(fv,path+xua_file_name);
   reset(fv,1);
   filemode := ab;
   if ioresult <> 0 then
      begin
        read_xuafile_to_xuaset_by_file(path+xua_file_name,ucrc,shareon);
        exit;
      end;

   recs := filesize(fv);
   if recs > memavail then
      begin
        close(fv);
        read_xuafile_to_xuaset_by_file(path+xua_file_name,ucrc,shareon);
        exit;
      end;

   GetMem(buf,recs);

   if buf = NIL then
      begin
        close(fv);
        read_xuafile_to_xuaset_by_file(path+xua_file_name,ucrc,shareon);
        exit;
      end;

   blockread(fv,buf^,recs,ab);
   close(fv);

{$IFDEF ADX}
   OpenXpAdxFile(path+adx_file_name);
{$ENDIF}

   recs := recs div sizeof(xua_rec);
   for ab := 1 to recs do
     begin
      with buf^[ab] do
        begin
          if integer(HIGHEST_AREANUM) < areanum then HIGHEST_AREANUM := areanum;
          if status <> -1 then
             if usrcrc = ucrc then
                begin
{$IFDEF ADX}
                  ia := areanum;
                  if Use_ADX then
                     begin
                       (* return the real area number (max, the index) *)
                       if not FindXpAdxAreaByCrc(areacrc,ia) then
                          begin
                            ia := -1;
                          end;
                     end;
                  if ia <> -1 then
                     begin
                       SetAreaBit(xua_set,ia);
                       if IsBitSet(_XUA_SYSOPSELECTED,xuaflags) then
                             SetAreabit(xsa_set,ia);
                     end;
{$ELSE}
                  SetAreaBit(xua_set,areanum);
                  if IsBitSet(_XUA_SYSOPSELECTED,xuaflags) then
                        SetAreabit(xsa_set,areanum);
{$ENDIF}
                end;
        end;
     end;
   FreeMem(buf,recs*sizeof(xua_rec));

{$IFDEF ADX}
   CloseXpADXFile;
{$ENDIF}

  end;

Procedure save_xuaset_to_xuafile_by_file(path : pathstr; ucrc:longint; shareon : boolean);
{ save XPUSER.XUA file }
var clpos,lpos : longint; i : word; err : integer;  f: byte; cnt : word;
    ax : integer;
    lk        : boolean;

label seek_another_record;

  Procedure WriteRecord;
    begin
      write(fv_xua,xua_rec);
    end;

  begin
   open_xua_file(path+xua_file_name,shareon);
   lpos  := 0;
   clpos := 0;
 {$IFDEF ADX}
   OpenXpAdxFile(path+adx_file_name);
 {$ENDIF}
   for i:= 0 to HIGHEST_AREANUM do
      begin
        if isAreaSelected(xua_set,i) then
           begin
seek_another_record:
            fillchar(xua_rec,sizeof(xua_rec),0);
            lpos := filepos(fv_xua);
            while (not eof(fv_xua))
                   and (xua_rec.status<>-1)
                    and (xua_rec.usrcrc <> ucrc) do
                begin
                 read(fv_xua,xua_rec);
                 inc(lpos);
                end;
            if eof(fv_xua) or (xua_rec.status=-1) or (xua_rec.usrcrc = ucrc) then
               begin
                if (lpos > 0) and
                    ((xua_rec.status=-1) or (xua_rec.usrcrc = ucrc)) then lpos := lpos-1;
                seek(fv_xua,lpos);
                xua_rec.status    := 0;
                xua_rec.areanum   := i;
                xua_rec.areacrc   := 0;
     {$IFDEF ADX}
                FindXpAdxCRCByArea(xua_rec.areacrc,i);
     {$ENDIF}
                xua_rec.xuaflags  := 0;
                if (xua_rec.usrcrc <> ucrc) and (xua_rec.status = -1)
                   then xua_rec.scanlimit := 0
                   else begin
                         if isAreaSelected(xsa_set,i) then
                            xua_rec.xuaflags := xua_rec.xuaflags or _XUA_SYSOPSELECTED;
                        end;
                xua_rec.usrcrc    := ucrc;
                ax := 0;
                if shareon then
                  begin
                   if filelock(filerec(fv_xua).handle,lockregion,lpos*sizeof(xua_rec),sizeof(xua_rec),ax) then
                       begin
                         WriteRecord;
                         if filelock(filerec(fv_xua).handle,unlockregion,lpos*sizeof(xua_rec),sizeof(xua_rec),ax) then;
                       end
                   else begin
                         goto seek_another_record;
                        end
                   end
                else WriteRecord;
                clpos := filepos(fv_xua);
               end;
           end
      end;
   seek(fv_xua,clpos);
   while not eof(fv_xua) do
     begin
      read(fv_xua,xua_rec);
      if xua_rec.status <> -1 then
         if xua_rec.usrcrc = ucrc then
            begin
              xua_rec.status := -1;
              seek(fv_xua,clpos);
              WriteRecord;
            end;
      inc(clpos);
     end;
   close(fv_xua);
 {$IFDEF ADX}
   CloseXpAdxFile;
 {$ENDIF}
  end;

Procedure save_xuaset_to_xuafile_by_mem(path : pathstr; ucrc:longint; shareon : boolean);
{ save XPUSER.XUA file }
var buf     : ^xuabuftype;
    recs    : longint;
    bufsize : word;
    ab      : word;
    fv      : file;
    err,cnt : integer;
    clpos,
    lpos,
    curpos  : longint;
    i       : word;
    lk      : boolean;

     procedure writerecord;
       begin
         buf^[curpos]   := xua_rec;
         inc(curpos);
       end;

  begin
   assign(fv,path+xua_file_name);
   ab := filemode;
   if shareon then filemode := $42; {_READWRITE+_DENYWRITE}
   cnt := 3000;
   repeat
     reset(fv,1);
     err := ioresult;
     case err of
       0  : begin end; (* OK *)
       2  : begin
              rewrite(fv,1);
              err := 0;
            end;
       5,
      167 : begin
             delay(10);
             dec(cnt);
            end;
      else  writeln('W error : ',err);
     end;
   until (err=0) or (cnt=0);
   filemode := ab;
   if (err<>0) and (cnt=0) then
      begin
        exit;
      end;

   recs := filesize(fv);
   if (recs > memavail) or (recs = 0) then
      begin
        close(fv);
        Save_xuaset_To_xuafile_by_file(path+xua_file_name,ucrc,shareon);
        exit;
      end;

   bufsize := recs;
   GetMem(buf,bufsize);

   if (buf = NIL) then
      begin
        close(fv);
        Save_xuaset_To_xuafile_by_file(path+xua_file_name,ucrc,shareon);
        exit;
      end;

   blockread(fv,buf^,bufsize,ab);

   recs   := bufsize div sizeof(xua_rec);

   lpos   := 1;
   clpos  := 1;
   curpos := 1;
 {$IFDEF ADX}
   OpenXpAdxFile(path+adx_file_name);
 {$ENDIF}


   for i:= 0 to HIGHEST_AREANUM do
      begin
        if isAreaSelected(xua_set,i) then
           begin
            fillchar(xua_rec,sizeof(xua_rec),0);
            lpos    := curpos;
            xua_rec := buf^[curpos];
            with xua_rec do
              while (curpos <= recs) and (status<>-1) and (usrcrc <> ucrc) do
                begin
                 inc(curpos);
                 inc(lpos);
                 xua_rec := buf^[curpos];
                end;

            if (curpos > recs) then
               begin
                 xua_rec.status    := 0;
                 xua_rec.areanum   := i;
                 xua_rec.usrcrc    := ucrc;
                 xua_rec.scanlimit := 0;
                 xua_rec.xuaflags  := 0;
                 xua_rec.areacrc   := 0;
     {$IFDEF ADX}
                 FindXpAdxCRCByArea(xua_rec.areacrc,i);
     {$ENDIF}
                 if isAreaSelected(xsa_set,i) then
                    xua_rec.xuaflags := xua_rec.xuaflags or _XUA_SYSOPSELECTED;
                 seek(fv,filesize(fv));
                 blockwrite(fv,xua_rec,sizeof(xua_rec));
               end
            else
            if (xua_rec.status=-1) or (xua_rec.usrcrc = ucrc) then
               begin
               xua_rec.status    := 0;
               xua_rec.areanum   := i;
               xua_rec.xuaflags  := 0;
               xua_rec.areacrc   := 0;
   {$IFDEF ADX}
               FindXpAdxCRCByArea(xua_rec.areacrc,i);
   {$ENDIF}
               if (xua_rec.usrcrc <> ucrc) and (xua_rec.status = -1)
                   then xua_rec.scanlimit := 0
                   else begin
                         if isAreaSelected(xsa_set,i) then
                            xua_rec.xuaflags := xua_rec.xuaflags or _XUA_SYSOPSELECTED;
                        end;
                xua_rec.usrcrc    := ucrc;
                writerecord;
                clpos := curpos;
               end;
           end;
      end;
   curpos := clpos;
   while curpos <= recs do
     begin
      xua_rec := buf^[curpos];
      inc(curpos);
      if xua_rec.status <> -1 then
         if xua_rec.usrcrc = ucrc then
            begin
              xua_rec.status := -1;
              curpos         := clpos;
              writerecord;
            end;
      inc(clpos);
     end;
   seek(fv,0);
   blockwrite(fv,buf^,bufsize,ab);
   close(fv);
   FreeMem(buf,bufsize);
 {$IFDEF ADX}
   CloseXPadxFile;
 {$ENDIF}
  end;

Procedure Delete_XuaFile_By_Mem(path : pathstr; ucrc:longint);
{ delete user entries in XPUSER.XUA file }
var buf     : ^xuabuftype;
    bufsize : word;
    ab      : word;
    fv      : file;
    err     : integer;
    curpos  : longint;

  begin
   assign(fv,path+xua_file_name);
   reset(fv,1);
   err := ioresult;
   if (err<>0) then exit;

   bufsize   := filesize(fv);
   if (bufsize > memavail) then
      begin
        close(fv);
        exit;
      end;

   GetMem(buf,bufsize);
   if (buf = NIL) then
      begin
        close(fv);
        exit;
      end;

   blockread(fv,buf^,bufsize,ab);
   for curpos := 1 to (ab div sizeof(xua_rec)) do
      if buf^[curpos].usrcrc = ucrc then buf^[curpos].status := -1;
   seek(fv,0);
   blockwrite(fv,buf^,ab);
   close(fv);
   FreeMem(buf,bufsize);
  end;


end.
